﻿// --------------------------------------------------------------------------------------
//
// FishFaceTX.CS                                                   Ulrich Müller 13.10.09
//                                                                 Aspire C# 2005 Express
// Klassenbibliothek für den Betrieb des ROBO TX Controllers im Online Modus
//     auf Basis der fischertechnik Library ftMscLib.DLL
//
// --------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace FishFaceTX {
  public enum Dev {
    Main = 0, Ext1, Ext2, Ext3, Ext4, Ext5, Ext6, Ext7, Ext8
  }
  public enum Mot {
    M1 = 0, M2, M3, M4
  }
  public enum Out {
    O1 = 0, O2, O3, O4, O5, O6, O7, O8
  }
  public enum Inp {
    I0 = -1, I1 = 0, I2, I3, I4, I5, I6, I7, I8, C1 = 8, C2, C3, C4
  }
  public enum Cnt {
    C1 = 0, C2, C3, C4
  }
  public enum Dir {
    Off = 0, Left = 1, Right = 2, On = 1
  }
  public class FishFaceException : ApplicationException {
    public FishFaceException() {
    }
    public FishFaceException(string message) : base(message) {
    }
    public FishFaceException(string message, Exception inner) : base(message, inner) {
    }
  }
  public class FishFace {
    #region ---- externe Declares --------

    // --- ftMscLib.DLL Deklarationen -------------------------------
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxGetLibVersionStr(StringBuilder sb, uint len);
    [DllImport("ftMscLib.dll")]
    private static extern uint GetAvailableComPorts(int selectMode);
    [DllImport("ftMscLib.dll")]
    private static extern uint EnumComPorts(uint idx, out string comStr, uint maxlen);
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxInitLib();
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxCloseLib();
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxOpenComDevice(string ComName, uint Baud, ref uint err);
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxCloseDevice(uint ftHandle);
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxStartTransferArea(uint ftHandle);
    [DllImport("ftMscLib.dll")]
    private static extern uint ftxStopTransferArea(uint ftHandle);

    [DllImport("ftMscLib.dll")]
    private static extern uint SetOutPwmValues(uint ftHandle,
      int devId, int outId, int duty);
    [DllImport("ftMscLib.dll")]
    private static extern uint SetFtMotorConfig(uint ftHandle,
      int devId, int MotorId, bool status);
    [DllImport("ftMscLib.dll")]
    private static extern uint SetOutMotorValues(uint ftHandle,
      int devId, int MotorId, int duty_p, int duty_m, bool brake);
    [DllImport("ftMscLib.dll")]
    private static extern uint SetMotorExConfig(uint ftHandle,
      int devId, int MotorId, int duty_p, int direction, int SlaveId, int pulsCount);
    [DllImport("ftMscLib.dll")]
    private static extern uint StopMotorExConfig(uint ftHandle,
      int devId, int MotorId, bool status);
    [DllImport("ftMscLib.dll")]
    private static extern uint StopMotorExConfig(uint ftHandle, int devId);

    [DllImport("ftMscLib.dll")]
    private static extern uint SetFtUniConfig(uint ftHandle,
      int devId, int ioId, int mode, bool digital);
    [DllImport("ftMscLib.dll")]
    private static extern uint SetFtCntConfig(uint ftHandle,
      int devId, int cntId, int mode);
    [DllImport("ftMscLib.dll")]
    private static extern uint GetInIOValue(uint ftHandle,
      int devId, int ioId, ref short Value, ref bool overrun);
    [DllImport("ftMscLib.dll")]
    private static extern uint GetInCounterValue(uint ftHandle,
      int devId, int cntId, ref short count, ref bool state);
    [DllImport("ftMscLib.dll")]
    private static extern uint SetOutCounterReset(uint ftHandle,
      int devId, int cntId);

    // --- Win API Deklarationen ----------------------------------------------------
    [DllImport("kernel32")]
    protected static extern void Sleep(uint mSek);
    [DllImport("kernel32")]
    protected static extern uint GetTickCount();
    [DllImport("user32")]
    protected static extern uint GetAsyncKeyState(uint vKey);
    protected const uint ESCAPE = 0x1B;
    #endregion

    private const uint txOK = 0x00000000;
    private const uint txPollInterval = 10;
    private const int  txWarten = 99;// ---- Bluetooth erfordert 99, USB geht auch mit 44
    private const int  modeVolt = 0;
    private const int  modeOhm = 1;
    private const int  modeDist = 3;
    
    private bool[,]    firstTime = new bool[9, 12];
    private int[,]     zielListe = new int[9, 4];
    private Thread     counterThread;
    
    private uint ftHandle;
    private uint errCode = txOK;
    private bool FOpen = false;
    private bool FNotHalt = false;
    private bool FDoEvents = true;

    public FishFace() {
      FOpen = false;
      FNotHalt = false;
      FDoEvents = false;
    }
    public bool NotHalt {
      get {
        return FNotHalt;
      }
      set {
        FNotHalt = value;
      }
    }
    public static string Version {
      get {
        return "5.0.0.0";
      }
    }
    public static string LibVersion {
      get{ 
      StringBuilder sb = new StringBuilder(null, 26);
      return ((ftxGetLibVersionStr(sb, 25) != 0) ? sb.ToString() : null);}
    }

    public void  OpenController(string ComName) {
      if (ftxInitLib() != txOK)
        throw (new FishFaceException("ControllerProblem.Open"));
      ftHandle = ftxOpenComDevice(ComName, 38400, ref errCode);
      if (errCode != txOK)
        throw (new FishFaceException("ControllerProblem.Open"));
      if (ftxStartTransferArea(ftHandle) != txOK)
        throw (new FishFaceException("ControllerProblem.Open"));
      for (int i = 0; i < 9; i++) {
        for (int j = 0; j < 12; j++) firstTime[i, j] = false;
        for (int j = 0; j < 4; j++)  zielListe[i, j] = 0;
      }
      FNotHalt = false;
      FOpen = true;
      FDoEvents = true;
      Thread.Sleep(txWarten*5);
      counterThread = new Thread(new ThreadStart(CounterLoop));
      counterThread.IsBackground = true;
      counterThread.Start();
    }
    private void CounterLoop() {
      int ziel;
      short count = 0;
      bool state = false;
      while (FOpen && !FNotHalt && (GetAsyncKeyState(ESCAPE) == 0)) {
        for (int i = 0; i < 9; i++) {
          for (int j = 0; j < 4; j++) {
            ziel=zielListe[i,j];
            if (ziel > 0) {
              GetInCounterValue(ftHandle, i, j, ref count, ref state);
              if (count >= ziel) {
                SetOutMotorValues(ftHandle, i, j, 0, 0, false);
                zielListe[i, j] = 0;
              }
            }
          }
        }
        Thread.Sleep(txWarten);
      }
      for (int i = 0; i < 9; i++)
        for (int j = 0; j < 4; j++) SetOutMotorValues(ftHandle, i, j, 0, 0, false);
      FNotHalt = true;
    }
    public void  CloseController() {
      if (FOpen) {
        if (ftxStopTransferArea(ftHandle) != txOK)
          throw (new FishFaceException("ControllerProblem.Close"));
        if (ftxCloseDevice(ftHandle) != txOK)
          throw (new FishFaceException("ControllerProblem.Close"));
        if (ftxCloseLib() != txOK)
          throw (new FishFaceException("ControllerProblem.Close"));
        FNotHalt = true;
        Thread.Sleep(txWarten);
        FOpen = false;
      }
    }

    public bool Finish() {
      return Finish(Inp.I0);
    }
    public bool Finish(Inp InputNr) {
      return Finish(Dev.Main, InputNr);
    }
    public bool Finish(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.Finish"));
      if (FNotHalt) return true;
      try {
        if (((int)InputNr >= 0) && GetInput(devId, InputNr)) return true;
      }
      catch (FishFaceException) {
        FOpen = false;
        throw (new FishFaceException("InterfaceProblem.Finish"));
      }
      return false;
    }
    public void Pause(int mSek) {
      uint sZeit = GetTickCount();
      do {
        if (FDoEvents) Application.DoEvents();
        if (FNotHalt) return;
        Sleep(txPollInterval);
      } while (GetTickCount() - sZeit < mSek);
    }
    
    public int  GetAnalog(Inp InputNr) {
      return GetAnalog(Dev.Main, InputNr);
    }
    public int  GetAnalog(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetAnalog"));
      short wert = 0;
      bool ok = false;
      if (!firstTime[(int)devId, (int)InputNr]) {
        if (SetFtUniConfig(ftHandle, (int)devId, (int)InputNr, modeOhm, false) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetAnalog"));
        }
        Sleep(txWarten);
        firstTime[(int)devId, (int)InputNr] = true;
      }
      if (GetInIOValue(ftHandle, (int)devId, (int)InputNr, ref wert, ref ok) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.GetInput"));
      }
      return wert;
    }
    public int  GetDistance(Inp InputNr) {
      return GetDistance(Dev.Main, InputNr);
    }
    public int  GetDistance(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetDistance"));
      short wert = 0;
      bool ok = false;
      if (!firstTime[(int)devId, (int)InputNr]) {
        if (SetFtUniConfig(ftHandle, (int)devId, (int)InputNr, modeDist, false) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetDistance"));
        }
        Sleep(txWarten);
        firstTime[(int)devId, (int)InputNr] = true;
      }
      if (GetInIOValue(ftHandle, (int)devId, (int)InputNr, ref wert, ref ok) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.GetDistance"));
      }
      return wert;
    }
    public bool GetInput(Inp InputNr) {
      return GetInput(Dev.Main, InputNr);
    }
    public bool GetInput(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetInput"));
      short wert = 0;
      bool ok = false;
      if ((int)InputNr < 8) {
        if (!firstTime[(int)devId, (int)InputNr]) {
          if (SetFtUniConfig(ftHandle, (int)devId, (int)InputNr, modeOhm, true) != txOK) {
            FOpen = false;
            throw (new FishFaceException("ControllerProblem.GetInput"));
          }
          Sleep(txWarten);
          firstTime[(int)devId, (int)InputNr] = true;
        }
        if (GetInIOValue(ftHandle, (int)devId, (int)InputNr, ref wert, ref ok) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetInput"));
        }
        return wert == 0 ? false : true;
      }
      else {
        if (!firstTime[(int)devId, (int)InputNr - 8]) {
          if (SetFtCntConfig(ftHandle, (int)devId, (int)InputNr - 8, 1) != txOK)
            throw (new FishFaceException("ControllerProblem.GetInput"));
          Sleep(txWarten); 
          firstTime[(int)devId, (int)InputNr - 8] = true;
        }
        if (GetInCounterValue(ftHandle, (int)devId, (int)InputNr - 8,
          ref wert, ref ok) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetInput"));
        }
        return ok;
      }
    }
    public bool GetTrack(Inp InputNr) {
      return GetTrack(Dev.Main, InputNr);
    }
    public bool GetTrack(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetTrack"));
      Int16 wert = 0;
      bool ok = false;
      if (!firstTime[(int)devId, (int)InputNr]) {
        if (SetFtUniConfig(ftHandle, (int)devId, (int)InputNr, modeVolt, true) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetTrack"));
        }
        Sleep(txWarten);
        firstTime[(int)devId, (int)InputNr] = true;
      }
      if (GetInIOValue(ftHandle, (int)devId, (int)InputNr, ref wert, ref ok) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.GetTrack"));
      }
      return wert > 0 ? false : true;
    }
    public int  GetVoltage(Inp InputNr) {
      return GetVoltage(Dev.Main, InputNr);
    }
    public int  GetVoltage(Dev devId, Inp InputNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetVoltage"));
      short wert = 0;
      bool ok = false;
      if (!firstTime[(int)devId, (int)InputNr]) {
        if (SetFtUniConfig(ftHandle, (int)devId, (int)InputNr, modeVolt, false) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.GetVoltage"));
        }
        Sleep(txWarten);
        firstTime[(int)devId, (int)InputNr] = true;
      }
      if (GetInIOValue(ftHandle, (int)devId, (int)InputNr, ref wert, ref ok) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.GetVoltage"));
      }
      return wert;
    }

    public int  GetCounter(Cnt Counter) {
      return GetCounter(Dev.Main, Counter);
    }
    public int  GetCounter(Dev devId, Cnt Counter) {
      if (!FOpen) throw (new FishFaceException("KeinOpen.GetCounter"));
      short wert=0;
      bool over=false;
      if (GetInCounterValue(ftHandle, (int)devId, (int)Counter, 
        ref wert, ref over) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.GetCounter"));
      }
      return wert;
    }
    public void ClearCounter(Cnt Counter) {
      ClearCounter(Dev.Main, Counter);
    }
    public void ClearCounter(Dev devId, Cnt Counter) {
      if (!FOpen) throw (new FishFaceException("KeinOpen.ClearCounter"));
      if (SetOutCounterReset(ftHandle, (int)devId, (int)Counter) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.ClearCounter"));
      }
    }
    
    public void SetLamp(Out LampNr, Dir OnOff) {
      SetLamp(Dev.Main, LampNr, OnOff, OnOff == Dir.On ? 512 : 0);
    }
    public void SetLamp(Dev devId, Out LampNr, Dir OnOff) {
      SetLamp(devId, LampNr, OnOff, OnOff == Dir.On ? 512 : 0);
    }
    public void SetLamp(Out LampNr, int Power) {
      SetLamp(Dev.Main, LampNr, Power > 0 ? Dir.On : Dir.Off, Power);
    }
    public void SetLamp(Dev devId, Out LampNr, int Power) {
      SetLamp(devId, LampNr, Power > 0 ? Dir.On : Dir.Off, Power);
    }
    public void SetLamp(Dev devId, Out LampNr, Dir OnOff, int Power) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.SetLamp"));
      if (SetOutPwmValues(ftHandle, (int)devId, (int)LampNr, 
        OnOff == Dir.On ? Power : 0) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.SetLamp"));
      }
    }
    
    public void SetMotor(Mot MotNr, Dir Direction) {
      SetMotor(Dev.Main, MotNr, Direction, Direction == Dir.Off ? 0 : 512);
    }
    public void SetMotor(Dev devId, Mot MotNr, Dir Direction) {
      SetMotor(devId, MotNr, Direction, Direction == Dir.Off ? 0 : 512);
    }
    public void SetMotor(Mot MotNr, Dir Direction, int Speed) {
      SetMotor(Dev.Main, MotNr, Direction, Speed);
    }
    public void SetMotor(Dev devId, Mot MotNr, Dir Direction, int Speed) {
      StartMotor(devId, MotNr, Direction, Speed, 0);
    }
    public void StartMotor(Mot MotNr, Dir Direction, int Speed, int ICount) {
      StartMotor(Dev.Main, MotNr, Direction, Speed, ICount);
    }
    public void StartMotor(Dev devId, Mot MotNr, Dir Direction, int Speed, int ICount) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.SetMotor"));
      int links;
      int rechts;
      switch (Direction) {
        case Dir.Left:
          links = Speed;
          rechts = 0;
          break;
        case Dir.Right:
          links = 0;
          rechts = Speed;
          break;
        default:
          links = 0;
          rechts = 0;
          break;
      }
      if (SetOutMotorValues(ftHandle, (int)devId, (int)MotNr,
        links, rechts, false) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.SetMotor"));
      }
      if (ICount > 0) {
        if (SetOutCounterReset(ftHandle, (int)devId, (int)MotNr) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.SetMotor"));
        }
        Thread.Sleep(txWarten);
        zielListe[(int)devId, (int)MotNr] = ICount;
      }
    }
    public void StartTwins(Mot MotNr, Mot SlaveNr,
      Dir Direction, int Speed, int ICount) {
      StartTwins(Dev.Main, MotNr, SlaveNr, Direction, Speed, ICount);
    }
    public void StartTwins(Dev devId, Mot MotNr, Mot SlaveNr, 
      Dir Direction, int Speed, int ICount) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.StartTwins"));
      if (SetMotorExConfig(ftHandle, (int)devId, (int)MotNr,
        Speed, Direction == Dir.Left ? 0 : 1, (int)SlaveNr, ICount) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.StartTwins"));
      }
      zielListe[(int)devId, (int)MotNr] = ICount;
      zielListe[(int)devId, (int)SlaveNr] = ICount;
      Thread.Sleep(txWarten*3);
    }
    public void StopTwins(Mot MotNr, Mot SlaveNr) {
      StopTwins(Dev.Main, MotNr, SlaveNr);
    }
    public void StopTwins(Dev devId, Mot MotNr, Mot SlaveNr) {
      if (FDoEvents) Application.DoEvents();
      if (!FOpen) throw (new FishFaceException("KeinOpen.StopTwins"));
      if (StopMotorExConfig(ftHandle, (int)devId) != txOK) {
        FOpen = false;
        throw (new FishFaceException("ControllerProblem.StopTwins"));
      }
      zielListe[(int)devId, (int)MotNr] = 0;
      zielListe[(int)devId, (int)SlaveNr] = 0;
      Thread.Sleep(txWarten);
    }

    public void WaitForMotors(params Mot[] MotorNrs) {
      WaitForMotors(Dev.Main, MotorNrs);
    }
    public void WaitForMotors(Dev devId, params Mot[] MotorNrs) {
      if (!FOpen) throw (new FishFaceException("KeinOpen.WaitForMotors"));
      Sleep(txWarten);
      int c = 0;
      do {
        Thread.Sleep(txWarten);
        c = 0;
        for (int i = 0; i < MotorNrs.Length; i++) {
          c += zielListe[i, (int)MotorNrs[i]];
        }
        if (c == 0) return;
        if (FDoEvents) Application.DoEvents();
      } while (!FNotHalt);
    }
    public void WaitForCount(Cnt CounterNr, int NrOfChanges) {
      WaitForCount(Dev.Main, CounterNr, NrOfChanges, Inp.I0);
    }
    public void WaitForCount(Cnt CounterNr, int NrOfChanges, Inp TermInputNr) {
      WaitForCount(Dev.Main, CounterNr, NrOfChanges, TermInputNr);
    }
    public void WaitForCount(Dev devId, Cnt CounterNr, int NrOfChanges) {
      WaitForCount(devId, CounterNr, NrOfChanges, Inp.I0);
    }
    public void WaitForCount(Dev devId, Cnt CounterNr, int NrOfChanges, Inp TermInputNr) {
      short count = 0;
      bool state = false;
      if (!FOpen) throw (new FishFaceException("KeinOpen.WaitForCount"));
      if (SetOutCounterReset(ftHandle, (int)devId, (int)CounterNr) != txOK)
        throw (new FishFaceException("ControllerProblem.WaitForCount"));
      Sleep(txWarten);
      do {
        if (GetInCounterValue(ftHandle, (int)devId, (int)CounterNr,
          ref count, ref state) != txOK) {
          FOpen = false;
          throw (new FishFaceException("ControllerProblem.WaitForCount"));
        }
        Sleep(txPollInterval);
        if (FNotHalt) return;
        if (TermInputNr > 0) if (GetInput(devId, TermInputNr)) return;
        if (FDoEvents) Application.DoEvents();
      } while (count < NrOfChanges);
    }
    public void WaitForInput(Inp InputNr) {
      WaitForInput(Dev.Main, InputNr, true);
    }
    public void WaitForInput(Dev devId, Inp InputNr) {
      WaitForInput(devId, InputNr, true);
    }
    public void WaitForInput(Inp InputNr, bool OnOff) {
      WaitForInput(Dev.Main, InputNr, OnOff);
    }
    public void WaitForInput(Dev devId, Inp InputNr, bool OnOff) {
      if (!FOpen) throw (new FishFaceException("KeinOpen.WaitForInput"));
      try {
        do {
          if (FNotHalt) return;
          Sleep(txPollInterval);
          if (FDoEvents) Application.DoEvents();
        } while (!GetInput(devId, InputNr) == OnOff);
      }
      catch (FishFaceException) {
        throw (new FishFaceException("InterfaceProblem.WaitForInput"));
      }
    }
    public void WaitForHigh(Inp InputNr) {
      WaitForHigh(Dev.Main, InputNr);
    }
    public void WaitForHigh(Dev devId, Inp InputNr) {
      WaitForInput(devId, InputNr, false);
      WaitForInput(devId, InputNr, true);
    }
    public void WaitForLow(Inp InputNr) {
      WaitForLow(Dev.Main, InputNr);
    }
    public void WaitForLow(Dev devId, Inp InputNr) {
      WaitForInput(devId, InputNr, true);
      WaitForInput(devId, InputNr, false);
    }
  }
}
